/*
 * Help Access Library 
 * A Library to access the contents of Windows Help files.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* SHGIMPOR.C -----------------------------------------
 * @doc
 * @module SHGIMPOR.C |
 *
 * Import-Support for SHG-Files.
 *
 * Converts SHG-Files (Segmented HyperGraphic) Files created
 * by Microsofts SHED or HC31 into DIB or METAFILE-Format
 *
 * Help Access Library Project.
 *
 *-----------------------------------------------------*/

#include <windows.h>
#include <windowsx.h>
#include <string.h>
#include <stdio.h>
#include "hlpacces.h"
#include "top.h"


#include "seg.h"
#include "shgimpor.h"           // SHG-Import API-Definition

DWORD doLZ   (LPBYTE object, LPBYTE Buffer, DWORD NumBytes);
DWORD doRLE  (LPBYTE object, LPBYTE Buffer, DWORD NumBytes);
DWORD doPLAIN(LPBYTE object, LPBYTE Buffer, DWORD NumBytes);
DWORD doLZRLE(LPBYTE object, LPBYTE Buffer, DWORD NumBytes);

/* Structures for .SHG files */

/* SHG File Header */
#pragma pack(1)
typedef struct tagSHGFILEHEADER 
{
  char  sfType[2];    /* Must be 'lp' or 0x706C     */
                        /* Inside of .HLP-Files also 'lP' or 0x506C, when Data is LZ-Compressed */
  LEWORD  sfNumObjects; /* Number of objects in file  */
} PACKED SHGFILEHEADER;


/* SHG Image Header */
#ifdef NEIN
typedef struct tagSHGIMAGEHEADER {
    LEWORD  siImageType;   /* 0x0106=.BMP  0x108=.WMF           */
                           /* 0x0206=.BMP  with LZ77-Compressed Data (Inside of .HLP-Files) */
    BYTE    siDPI;         /* Dots Per Inch x 2 (0x10 for .wmf) */
}
PACKED SHGIMAGEHEADER;
#endif
#pragma pack()

/* Defines for image type and monitor type */
#define IT_BMP_PLAIN 0x0006
#define IT_BMP_RLE   0x0106
#define IT_BMP_LZ    0x0206
#define IT_BMP_LZRLE 0x0306

#define IT_WMF_PLAIN 0x0008
#define IT_WMF_RLE   0x0108
#define IT_WMF_LZ    0x0208

/* SHG Bitmap Header */
#pragma pack(1)
typedef struct tagSHGBITMAPHEADER 
{
    DWORD sbDPIX;        /* Dots Per Inch Y-Direction     */
    DWORD sbDPIY;        /* Dots Per Inch X-Direction     */
    UINT  sbPlanes;      /* Number of Planes in Bitmap    */
    UINT  sbNumBits;     /* Number bits per pixel x 2     */
    DWORD sbWidth;       /* Width x 2                     */
    DWORD sbHeight;      /* Height x 2                    */
    DWORD sbNumQuads;    /* Number RGB Quads x 2          */
    UINT  sbNumImp;      /* Number of 'important' RGB Qds */
    DWORD sbCmpSize;     /* Size of Compressed BMP x 2    */
    DWORD sbSizeHS;      /* Size of Hotspot Data area x 2 */
    DWORD sbOffsetBM;    /* Offset from beginning of Structure to Compressed Picture Data */
    DWORD sbSizeImage;   /* size ImageHdr+BmpHdr+ImageDat == Offset to Hotspots */
} SHGBITMAPHEADER;

/* SHG Metafile Header */
typedef struct tagSHGMETAFILEHEADER {
    BYTE  smAlwaysTen; /* Always 0x10                       */
    WORD  smXWidth;    /* Width of image in metafile units  */
    WORD  smYHeight;   /* Height of image in metafile units */
//    WORD  smUnk1;
    DWORD smUnk1;      /* Changed Bernd Herd 21.6.95 */
    DWORD smCmpSize;   /* Size of compressed metafile    */
    DWORD smSizeHS;    /* Size of hot spot data area x 2 */
    DWORD smOffsetWM;  /* Offset from beginning of Structure to Compressed Picture Data - 4 Usually 21, in users\halser\invalid_shg.zip 23. Changed Bernd 14.7.98 */
    DWORD smSizeImage; /* Size ImageHdr+wmfHdr+ImageDat  */
} SHGMETAFILEHEADER;

/* Hot Spot Header */
typedef struct tagHOTSPOTHEADER 
{
    BYTE    hhOne;           /* Always 0x01                                */
    LEWORD  hhNumHS;         /* Number of Hot Spots                */
    LEDWORD hhContextOffset; /* Offset to Cntxt Strings & IDs*/
} HOTSPOTHEADER;

/* Hot Spot Record */
typedef struct tagHOTSPOTRECORD 
{
    LEWORD  hrType; /* Hot Spot Type. See below         */
    BYTE    hrZero; /* Always 0                                         */
    LEWORD  hrLeft; /* Bounding box of Hot Spot         */
    LEWORD  hrTop;
    LEWORD  hrRight;
    LEWORD  hrBottom;
    LEDWORD hrMacOffset;/* Offset to macro for Hot Spot */
} HOTSPOTRECORD;
#pragma pack()


#define HS_INVISJUMP            0x04E7
#define HS_INVISPOPUP           0x04E6
#define HS_INVISMACRO           0x04CC
#define HS_VISJUMP              0x00E3
#define HS_VISPOPUP             0x00E2
#define HS_VISMACRO             0x00C8

/* Flags an incorrect value for known fields */
#define CheckUVal(a,b) \
       {if (a != b) {  \
       HlpDisplayWarning("UINT: Not a 0x%04X?\n\n 0x%04x\n",b, a); exit(1); } }

#define CheckBVal(a,b) \
       {if (a != b) {  \
       HlpDisplayWarning("BYTE: Not a 0x%02X?\n\n 0x%02x\n",b, a); exit(1); } }

#define ReadString(f, s) { char *p = (char *)(s); \
                  while ((*p++ = *f++) != 0) ; *p = 0; }

/* Reads in a BYTE. If it's odd, it reads in a WORD. Returns 1/2 value */
WORD ReadWordVal(LPBYTE *object)
{
    BYTE a, b;

    b = 0;
    a=*(*object)++;
    if (a % 2)
        b=*(*object)++;

    return (WORD) ((WORD)b*256 + a) / (WORD)2;
}

/* Reads in a WORD. If it's odd, it reads in a DWORD. Returns 1/2 value */
DWORD ReadDWordVal(LPBYTE *object)
{   WORD a, b;

    b = 0;
    a=iGetLeWordValue(object, NULL);

    if (a % 2)
    {
       b=iGetLeWordValue(object, NULL);
    }

    return (DWORD) ((DWORD)b*65536L + a) /2;
}



/* Dumps HotSpot data from .SHG file */
void DumpHotSpots(
                #ifdef _Windows
                HWND         hwParent,                  // Parent-Window for error Messages
                #endif
                LPBYTE       object,                    // Pointer to start of this Object
                HGLOBAL *lphSegInfo,                    // Segmentation Informations
                SHGBITMAPHEADER *SHGBM)                 // Bitmap heade to retreive width and height for Bitmaps

{   HGLOBAL         hgSeg = GlobalAlloc(GHND, 0x8000L);
    LPSEGINFOHEADER sih   = (LPSEGINFOHEADER) GlobalLock(hgSeg);
    LPSEGJUMP       sj    = (LPSEGJUMP) (sih+1);

    HOTSPOTHEADER   *HSHead=((HOTSPOTHEADER   *)object)++;
    LPBYTE          strt       =         object;
    LPSTR           context    = (LPSTR)(object+
                                        leword(HSHead->hhNumHS)*sizeof(HOTSPOTRECORD)+
                                        ledword(HSHead->hhContextOffset));
    HOTSPOTRECORD   *HSRec;
    LPSTR           String;
    UINT            i, len;

    // Initialize...
    sih->Magic              =SEG_MAGIC;
    sih->SegJumpCount       =leword(HSHead->hhNumHS);
    sih->HeaderAndStringSize=sizeof(SEGINFOHEADER) + sih->SegJumpCount * sizeof(SEGJUMP);
    if (SHGBM)
      {
        sih->Width = SHGBM->sbWidth;
        sih->Height= SHGBM->sbHeight;
      }

    for(i = 0; i < leword(HSHead->hhNumHS); i++, sj++)
    {   HSRec = ((HOTSPOTRECORD   *) object)++;

        switch(leword(HSRec->hrType)) 
          {
        case HS_VISJUMP:
        case HS_INVISJUMP:  sj->flags = SI_JUMP;  break;
        case HS_VISPOPUP:
        case HS_INVISPOPUP: sj->flags = SI_POPUP; break;
        case HS_VISMACRO:
        case HS_INVISMACRO: sj->flags = SI_MACRO; break;
        default:
#           ifdef _Windows
            MessageBox(hwParent, "Unknown Hotspot Type", "SHG-Import", MB_OK);
#           else
            printf("Unknown Hotspot Type\n");
#           endif
        }
        sj->left  =leword(HSRec->hrLeft);
        sj->right =leword(HSRec->hrLeft)+leword(HSRec->hrRight);
        sj->top   =leword(HSRec->hrTop);
        sj->bottom=leword(HSRec->hrTop)+leword(HSRec->hrBottom);

        //----------- Get Macro- or Contextstring ----------------------------
        if (sj->flags & SI_MACRO)
            String=(LPSTR)     (strt+
                                leword(HSHead->hhNumHS)*sizeof(HOTSPOTRECORD)+
                                ledword(HSRec->hrMacOffset));
         else
            String=context+strlen(context)+1;

        sj->macro.dw = sih->HeaderAndStringSize;
        len=strlen(String);
        memcpy(((LPSTR) sih) + sih->HeaderAndStringSize, String, len+1);
                                sih->HeaderAndStringSize    +=    len+1;

        //----------- Get Hotspot ID -----------------------------------------
        String=context;

        sj->HotspotId.dw = sih->HeaderAndStringSize;
        len=strlen(String);
        memcpy(((LPSTR) sih) + sih->HeaderAndStringSize, String, len+1);
                                sih->HeaderAndStringSize    +=    len+1;

        context+=strlen(context)+1;
        context+=strlen(context)+1;
    }

    GlobalUnlock(hgSeg);

    *lphSegInfo=hgSeg;
}


/* Reads in a bitmap header. Doubles the
   size of certain fields when necessary. */
LPBYTE ReadBMHeader(LPBYTE object, SHGBITMAPHEADER *SHGBM)
{
    /* Read first three fields unmodified */
    SHGBM->sbDPIX    = ReadDWordVal(&object); /* May be in "Dots Per Inch" or "Pixel per Meter" */
    SHGBM->sbDPIY    = ReadDWordVal(&object);
    SHGBM->sbPlanes  = ReadWordVal(&object);
    SHGBM->sbNumBits = ReadWordVal(&object);
    SHGBM->sbWidth   = ReadDWordVal(&object);
    SHGBM->sbHeight  = ReadDWordVal(&object);
    SHGBM->sbNumQuads= ReadDWordVal(&object);
    SHGBM->sbNumImp  = (WORD) ReadDWordVal(&object);
    SHGBM->sbCmpSize = ReadDWordVal(&object);
    SHGBM->sbSizeHS  = ReadDWordVal(&object);
    SHGBM->sbOffsetBM = iGetLeDWordValue(&object, NULL);
    //printf("sbOffsetBM=%d,", SHGBM->sbOffsetBM);
    SHGBM->sbSizeImage = iGetLeDWordValue(&object, NULL);
    //printf("sbSizeImage=%d", SHGBM->sbSizeImage);
    //    memcpy(&SHGBM->sbOffsetBM, object, 8);
//    memcpy(&SHGBM->sbSizeImage, object+4, 4);
    //object+=8;                // first 4 unknown...
    return object;
}

/* Build the bitmap file from the .SHG file */
BOOL BuildBitmap(
                #ifdef _Windows
                HWND         hwParent,                  // Parent-Window for error Messages
                #endif
                LPBYTE       object,                    // Pointer to start of this Object
                int          siImageType,
                SHGBITMAPHEADER *SHGBM,
                HGLOBAL *lphDib)            // Segmentation Informations
{   HGLOBAL          hMem;
    LPBYTE           pMem;
    BITMAPINFOHEADER bmih;
    int              sizeofheader;

    memset(&bmih, 0, sizeof(bmih));
    bmih.biSize   = sizeof(bmih);
    bmih.biWidth  = SHGBM->sbWidth;
    bmih.biHeight = SHGBM->sbHeight;
    bmih.biPlanes = 1;
    bmih.biBitCount = SHGBM->sbNumBits;
    bmih.biCompression = BI_RGB;
    bmih.biSizeImage = 0; /* We'll fill this in later */
    bmih.biXPelsPerMeter =
    bmih.biYPelsPerMeter = (SHGBM->sbDPIX) * 40L;
    bmih.biClrUsed = SHGBM->sbNumQuads;
    bmih.biClrImportant = 0;

    //-------- Let's allocate a fairly big enoug buffer -------------
    sizeofheader = sizeof(bmih) +
                   (WORD) SHGBM->sbNumQuads * sizeof(RGBQUAD);

    hMem = GlobalAlloc(GHND, sizeofheader +
                             ((bmih.biWidth * bmih.biBitCount + 31) / 32 * 8) * bmih.biHeight + 100);
    if (!hMem)
    {  
#      ifdef _Windows
       MessageBox(hwParent, "Not enough memory", "SHG-Import", MB_SYSTEMMODAL);
#      else
       printf("Not enough memory\n");
#      endif
       return FALSE;
    }

    pMem = (LPBYTE) GlobalLock(hMem);
    memcpy(pMem             , &bmih , sizeof(bmih));
    memcpy(pMem+sizeof(bmih), object, SHGBM->sbNumQuads * sizeof(RGBQUAD));
    object += SHGBM->sbNumQuads * sizeof(RGBQUAD);

    switch (siImageType) 
    {
        case IT_BMP_PLAIN: bmih.biSizeImage = doPLAIN(object, pMem+sizeofheader, SHGBM->sbCmpSize); break;
        case IT_BMP_RLE:   bmih.biSizeImage = doRLE  (object, pMem+sizeofheader, SHGBM->sbCmpSize); break;
        case IT_BMP_LZ:    bmih.biSizeImage = doLZ   (object, pMem+sizeofheader, SHGBM->sbCmpSize); break;
        case IT_BMP_LZRLE: bmih.biSizeImage = doLZRLE(object, pMem+sizeofheader, SHGBM->sbCmpSize); break;
    }

    GlobalUnlock(hMem);

    *lphDib=hMem;

    return TRUE;
}

/*----------------------------------------------------------
 * Convert Bitmap-Object from SHG-File into a DIB
 *
 *  Returns: TRUE if normal .BMP-File
 */
BOOL BitMapDump(
                #ifdef _Windows
                HWND         hwParent,                  // Parent-Window for error Messages
                #endif
                LPBYTE       object,                    // Pointer to start of this Object
                int          siImageType,
                HGLOBAL *lphDib,                    // DIB-Handle
                HGLOBAL *lphSegInfo)                // Segmentation Informations
{
    SHGBITMAPHEADER SHGBM;
    BOOL            ok;

    object = ReadBMHeader(object, &SHGBM);

    ok = lphDib!=NULL ? BuildBitmap(
#                           ifdef _Windows
                            hwParent, 
#                           endif
                            object, 
                            siImageType, 
                            &SHGBM, 
                            lphDib)
                      : TRUE;   

    if (SHGBM.sbSizeHS && lphSegInfo)
       DumpHotSpots(
#          ifdef _Windows
           hwParent, 
#          endif
           object+SHGBM.sbNumQuads * sizeof(RGBQUAD) + SHGBM.sbCmpSize,
           lphSegInfo,
           &SHGBM);

    return ok;
}


/* Reads in a metafile header. Doubles the
   size of certain fields when necessary. */
LPBYTE ReadWMHeader(LPBYTE object, SHGMETAFILEHEADER *SHGWM)
{
  /* Read the binary Metafile-header into our byteorder-independent structure */
  SHGWM->smAlwaysTen = *object++;
  SHGWM->smXWidth = object[0] | object[1]<<8; object+=2;
  SHGWM->smYHeight = object[0] | object[1]<<8; object+=2;

  SHGWM->smUnk1    = ReadDWordVal(&object);
  SHGWM->smCmpSize = ReadDWordVal(&object);
  SHGWM->smSizeHS  = ReadDWordVal(&object);

  SHGWM->smOffsetWM= object[0] | object[1]<<8 | object[2]<<16 | object[3]<<24; object+=4;
  SHGWM->smSizeImage= object[0] | object[1]<<8 | object[2]<<16 | object[3]<<24; object+=4;
  
  return object;
}






/* Converts the regular metafile to a placeable metafile */
void WritePlaceableHeader(LPMETAFILEHEADER WMFHead, SHGMETAFILEHEADER *SHGWM)
{
    memset(WMFHead, 0, sizeof(METAFILEHEADER));

    WMFHead->key = 0x9AC6CDD7L; /* file identifier */
    WMFHead->hmf = 0;
    WMFHead->inch = 254;
    WMFHead->reserved = 0;
    WMFHead->bbox.top    =
    WMFHead->bbox.left   = 0;
    WMFHead->bbox.right  = SHGWM->smXWidth / 10;
    WMFHead->bbox.bottom = SHGWM->smYHeight/ 10;
    WMFHead->checksum = 0x0000 ^ 0xCDD7 ^ 0x9AC6 ^ WMFHead->hmf ^
    WMFHead->bbox.left ^ WMFHead->bbox.top ^ WMFHead->bbox.right ^
    WMFHead->bbox.bottom ^ WMFHead->inch;
}



/**
 * Create an in-memory representation of a Metafile stored in a .shg
 *
 * @param hwParent         Parent Window Handle (only used for M$-Windows)
 * @param object           Pointer to the start of this object in the .SHG file
 * @param siImageType      Image-Type value (Encodes compression method)
 * @param SHGWM            SHG Metafile Header read from the image
 * @param lpMetaFile       Pointer to an handle that will receive the resulting in-memory representatzion
 * @param piImageSize      Pointer to an Integer that will contain the size of the decompressed Metafile in lpMetaFile on return
 * @param lpNetaFileHeader Pointer to a structure that will be filled with Informations on placeable Metafiles
 *
 * @return TRUE on Success
 */
BOOL BuildWMF(  
                LPBYTE       object,
                int          siImageType,
                SHGMETAFILEHEADER *SHGWM,
                HMETAFILE*   lphMetaFile,
                int*             piImageSize,
                LPMETAFILEHEADER lpMetaFileHeader)
{   DWORD   ImageSize;
    HGLOBAL hMem=GlobalAlloc(GHND, SHGWM->smCmpSize * 50);
    LPBYTE  pMem=(LPBYTE) GlobalLock(hMem);

    /* Step 1 decompress file to temp file */
    switch (siImageType)
    {
        case IT_WMF_PLAIN: ImageSize = doPLAIN(object, pMem, SHGWM->smCmpSize); break;
        case IT_WMF_RLE:   ImageSize = doRLE  (object, pMem, SHGWM->smCmpSize); break;
        case IT_WMF_LZ:    ImageSize = doLZ   (object, pMem, SHGWM->smCmpSize); break;
//      case IT_WMF_LZRLE: ImageSize = doLZRLE(object, pMem, SHGWM->smCmpSize); break;
    }

    GlobalUnlock(hMem);
    hMem=GlobalReAlloc(hMem, ImageSize, GHND);

#   ifdef __WIN32__
    pMem=(LPBYTE) GlobalLock(hMem);
    *lphMetaFile = SetMetaFileBitsEx(ImageSize, pMem);
    GlobalUnlock(hMem);
    GlobalFree(hMem);
#   else
    *lphMetaFile = (HMETAFILE) hMem;
#   endif
    *piImageSize = ImageSize;

    return TRUE;
}




/* Dump the .WMF data from the .SHG file
   Returns: TRUE if normal .WMF-File
*/

BOOL WMFDump(   
                #ifdef _Windows
                HWND         hwParent,                  // Parent-Window for error Messages
                #endif
                LPBYTE       object,                    // Pointer to start of this Object
                LPBYTE       shg,                       // Equals FileStartPos
                int          siImageType,
                HMETAFILE*   lphMetaFile,           // Metafile-Handle for SHG containing WMF
                int*             piImageSize,
                LPMETAFILEHEADER lpMetaFileHeader,      // Will be filled with Informations on placeable Metafiles
                HGLOBAL *lphSegInfo)                // Segmentation Informations
{   SHGMETAFILEHEADER SHGWM;
    BOOL            Build;

    object = ReadWMHeader(object, &SHGWM);
    
    if (lpMetaFileHeader!=NULL)
      {
        WritePlaceableHeader(lpMetaFileHeader, &SHGWM);
      }

    Build = lphMetaFile!=NULL ? BuildWMF(
                                  object, 
                                  siImageType, 
                                  &SHGWM, 
                                  lphMetaFile, 
                                  piImageSize,
                                  lpMetaFileHeader)
                              : TRUE;

//    fseek(SHGFile, (29+SHGWM.smCmpSize)+FileStartPos, SEEK_SET);
    if (SHGWM.smSizeHS && lphSegInfo)
      {
        DumpHotSpots(
#ifdef _Windows
                     hwParent, 
#endif
                     shg+(SHGWM.smOffsetWM+29-21+SHGWM.smCmpSize), 
                     lphSegInfo,
                     NULL);
      }

    return Build;
}





/* -----------------------------------------------
 * @func BOOL | SHGImport |
 *
 * Does convert a Memory-Block that signifies an SHG-File into
 * a DIB-Bitmap or Meatfile and Segmentation-Information in SEG-File
 * Format. The segmented Hypergraphics in SEG-File format is
 * documented in file SEG.H
 *
 * @comm Segmented Hypergraphic file can either be created based
 *       on a Bitmap (DIB) or on a Metafile. Both can be converted one
 *       into the other with some restrictions. This conversion code
 *       is not itself part of Help Access Library and can be found
 *       in our DaVinci package. Be aware that many Metafiles are
 *       just wrappers around Bitmaps and using PlayMetafile to
 *       transfer these into a Bitmap is NOT the best method of handling.
 * @xref <f SHGExport>
 *
 * @rvalue TRUE | On Successfull Execution
 * @rvalue FALSE| On Errors
 *------------------------------------------------*/

BOOL WINAPI SHGImport(
                #ifdef _Windows
                HWND             hwParent,        // @parm Parent-Window for error Messages
                #endif
                LPVOID           SHGData,         // @parm Input Data Block
                HGLOBAL *    lphDib,              // @parm Pointer to a DIB-Handle variable. Should be
                                                  //       Initialized by the application to NULL and will
                                                  //       Contain a DIB-Handle on return if the SHG-File contained a DIB
                HMETAFILE*   lphMetaFile,         // @parm Metafile-Handle for SHG containing WMF
                                                  //       Initialized by the application to NULL and will
                                                  //       Contain a HMETAFILE-Handle on return if the SHG-File contained a Metafile
                int *        piImageSize,         // @parm Out: Size of the memory block representing a metafile (lphMetaFile)
                LPMETAFILEHEADER lpMetaFileHeader,// @parm Will be filled with Informations on placeable Metafiles
                                                  //       This member is used only if the SHG-File has been built
                                                  //       based on a Metafile.
                HGLOBAL *    lphSegInfo           // @parm Segmentation Informations in format documented in SEG.H
                )
{   WORD                siImageType;
    int                 i;
    SHGFILEHEADER *     SHGHead = (SHGFILEHEADER *) SHGData;
    LPBYTE              shg     = ((LPBYTE) SHGData),
                        object;
    LEDWORD *           lpledwObjectOffsets = (LEDWORD *) (SHGHead+1);
    BOOL                rtn=FALSE;

    /* Make sure it's a .SHG or .MRB file */
    if (memcmp(SHGHead->sfType, "lp", 2) &&
        memcmp(SHGHead->sfType, "lP", 2)   )
        { 
#         ifdef _Windows
          MessageBox(hwParent, "Not a valid SHG-File", "SHG-Import", MB_OK | MB_ICONSTOP);
#         else
          printf("Not a valid SHG-File\n");
#         endif
          return FALSE;
        } else
//      for(i=0; i < leword(SHGHead->sfNumObjects); i++)
        {  i=0;

           /* Access object */
           object      = shg+ledword(lpledwObjectOffsets[i]);

           /* Read in Image Header */
           //           siImageType = *((WORD *) object)++;
           siImageType =      object[0] + 256 * object[1];
                              object+=2;

           switch (siImageType)
           {
            case IT_BMP_LZRLE:
            case IT_BMP_PLAIN:
            case IT_BMP_RLE:
            case IT_BMP_LZ:                        // Only valid inside of .HLP-Files
                       rtn = BitMapDump(
                                 #ifdef _Windows
                                 hwParent,
                                 #endif
                                 object,
                                 siImageType,
                                 lphDib,
                                 lphSegInfo);
                       break;

            case IT_WMF_PLAIN:  // Uncompressed Metafile Image
            case IT_WMF_RLE:    // RLE-Compressed Metafile Image
            case IT_WMF_LZ:     // LZ-Compressed Metafile Image
                       rtn = WMFDump(      
#                                          ifdef _Windows
                                           hwParent,
#                                          endif
                                           object,
                                           shg,
                                           siImageType,
                                           lphMetaFile,
                                           piImageSize,
                                           lpMetaFileHeader,
                                           lphSegInfo);
                       break;
            default:
#              ifdef _Windows
               MessageBox(hwParent, "Unknown Sub-Filetype", "SHG-Import", MB_OK | MB_ICONSTOP);
#              else
               printf("Unknown Sub-Filetype\n");
#              endif
           }
       }

       return rtn;
}




/*************************************************
  Returns a 1 if the bit is set, else 0 returned
**************************************************/

short static LzBitSet(BYTE BitMap, short Bit) {

   if (BitMap & (1<<Bit)) return 1;
                     else return 0;

}


/*************************************************
  Decides how many bytes to read, depending on the
  number of bits set in the Bitmap
**************************************************/

short static BytesToRead(BYTE BitMap)
{
    short TempSum, counter;

    TempSum = 8;
    for (counter = 0; counter < 8; counter ++)
       TempSum += LzBitSet(BitMap, counter);

    return TempSum;
}


/*************************************************
  Decompresses the data using Microsoft's LZ77
derivative.
**************************************************/
DWORD doLZ(LPBYTE object, LPBYTE Buffer, DWORD CompSize)
{
   DWORD  InBytes = 0;        /* How many bytes read in                    */
   DWORD  OutBytes = 0;       /* How many bytes written out                */
   BYTE   BitMap, Set[16];    /* Bitmap and bytes associated with it       */
   int    NumToRead;          /* Number of bytes to read for next group    */
   UINT   counter, Index;     /* Going through next 8-16 codes or chars    */
   DWORD  Length, Distance;   /* Code length and distance back in 'window' */
   LPBYTE CurrPos;            /* Where we are at any given moment          */
   LPBYTE CodePtr;            /* Poshorter to back-up in LZ77 'window'       */

   CurrPos = Buffer;

   /* Go through until we're done */
   while (InBytes < CompSize)
   {
      /* Get BitMap and data following it */
      BitMap = *object++; InBytes++;
      NumToRead = BytesToRead(BitMap);

      /* If we're trying to read more than we've got left, only read what we have left. */
      NumToRead = (int) ((CompSize - InBytes) < NumToRead ? CompSize-InBytes : NumToRead);

      memcpy(&Set, object, NumToRead);
                    object+=NumToRead;  
      InBytes += NumToRead;

      /* Go through and decode data */
      for (counter = 0, Index = 0; NumToRead > 0; counter++)
      {
         /* It's a code, so decode it and copy the data */
         if (LzBitSet(BitMap, counter)) {
            Length = ((Set[Index+1] & 0xF0) >> 4) + 3;
            Distance = (256 * (Set[Index+1] & 0x0F)) + Set[Index] + 1;
            NumToRead -= 2;

            CodePtr = CurrPos - Distance;

            /* Copy data from 'window' */
            while (Length) {
               *CurrPos++ = *CodePtr++;
               OutBytes++;
               Length--;
            }

            Index += 2;

         } /* if */

         else {
            *CurrPos++ = Set[Index++];
            NumToRead --;
            OutBytes++;
         }

      } /* for */

   } /* while  */

   return OutBytes;
}







/* Performs RLE decompression for bitmaps and metafiles */
DWORD doRLE(LPBYTE object, LPBYTE Buffer, DWORD NumBytes)
{
    DWORD i=1, imagesize=0;
    BYTE  j, count, data;

    do {
        count=*object++;
        i++;

        /* If the 8th bit is set, lower 7 bits has # bytes to copy */
        if (count & 0x80) {
            count -=0x80;
            imagesize+=count;
            while (count--) {
                *Buffer++=*object++;
                i++;
            }
        }
        /* If 8th bit not set, it's an RLE count */
        else {
            count = count & ((BYTE) 0x7F);
            data  = *object++;
            i++;
            /* Write out uncompressed */
            for (j = 1; j <= count; j++, imagesize++)
                *Buffer++=data;
        }
    }
    while(i < NumBytes);

    return imagesize;
}


/* Performs PLAIN Copy for bitmaps and metafiles */
DWORD doPLAIN(LPBYTE object, LPBYTE Buffer, DWORD NumBytes)
{   memcpy(Buffer, object, NumBytes);
    return NumBytes;
} // doPLAIN



DWORD doLZRLE(LPBYTE object, LPBYTE Buffer, DWORD NumBytes)
{        HGLOBAL hbuf2;
         LPBYTE  pbuf2;
         DWORD RleSize  = doLZ(object, Buffer, NumBytes),
               FullSize;        

         hbuf2 = GlobalAlloc(GHND, RleSize);
         pbuf2 = (LPBYTE) GlobalLock(hbuf2);
         memcpy(pbuf2, Buffer, RleSize);

         FullSize = doRLE(pbuf2, Buffer, RleSize);
         GlobalUnlock(hbuf2);
         GlobalFree(hbuf2);
         return FullSize;
} // doLZRLE




/*
 * Local variables:
 *  tab-width: 8
 * End:
 */
