/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* wc_html.c -- part of WinHelpCGI

   This piece enumerates a topic of a HLP file and does the
   conversion into the HTML file format.
 
   (C)1996-2004 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/


#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <locale.h>
#include "hlpacces.h"           // Help Access Library Functions
#include "shgimpor.h"           // SHG Import functions (for image maps)
#include "../libhlpaccess/seg.h"
#include "winhelpcgi.h"         // Header-File to connect the pieces of WinHelpCGI application

#ifdef ENABLE_NLS
#include <libintl.h>

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

#else
#define _(String) (String)
#endif

/// Informations stored for every single tabstop
typedef struct
{
  /// Tabstop positions
  int tabpos;

  /// Tabstop alignment
  int tabalign;
} TABSTOP;


/// Maximum nunber of tabstops to be stored
#define MAX_TABSTOPS 100

/// Informations stored for a set of tabstops
typedef struct
{
  TABSTOP  tabstop[MAX_TABSTOPS];       /* Tabstop positions and sizes */
  int      tabcount;                    /* Number of tabstops stored */
} TABSETTINGS;


/** Informations stored on images stored by numbers
 *
 * We try to access every image only once, saving width, height and the usemap flag.
 */
typedef struct
{
  int width;
  int height;
  BOOL usemap;
  BOOL tested;
} IMAGEINFO;


#define MAX_IMAGES 1000


/* The ENUMERATTION_PRIVATE_DATA structur contains fields that are needed by
   the callback-function during enumeration of a Topic. */
typedef struct
{
  BOOL     BrowseBackCalled;            /* TRUE after Title has been enumerated first */
  LPCSTR   HlpFileName;                 /* Filename of Help-File currently processed */
  DWORD    TopicOffset;                 /* TopicOffset of Topic currently processed */
  
  BOOL     Hidden;                      /* text is currently hidden (RTF \v)*/
  BOOL     InAnchor;                    /* An anchor-Tag <a ...> has is currently open */
  BOOL	  HtmlTable;                   /* A table-Tag <table> is currently open */
  BOOL	  HtmlRow;                     /* A Table-Row <tr><td> is currently open */
  BOOL	  RtfTable;                    /* RTF opened table with \intbl (will be closed by \row)
														 needed to collect several lines of a RTF-Tables to ONE
														 HTML-Table     
													 */
  BOOL	  InBrdrl;                     /* Border Right set (Display line on the left) */
  BOOL	  InBrdrr;                     /* Border Right set (Display line on the right) */
  BOOL	  InBrdrt;                     /* Border Top set (Display line above) */
  BOOL	  InBrdrb;                     /* Border Bottom set (Display line below) */

  BOOL     FontSet;                     /* A font has been set */
  
  LOGFONT  lf[MAX_FONTS];               /* Font-Name from RTF \fonttbl */
  LOGFONT  lflast;                      /* Last font set by <font...> tag */
  LOGFONT  lfnow;                       /* Font that need to be set for next Character */
  RGBQUAD  cl[MAX_FONTS];               /* Color-Informations from RTF \colortbl */
  RGBQUAD  cllast;                      /* Last color set b< <font...> tag */
  RGBQUAD  clnow;                       /* Color that need to be set for next Character */

  HIFS     ifs;                         /* Access to internal file System */
  BOOL     InPara;                      /* Paragraph has been opened */
  BOOL     InNsr;                       /* Currently in a non-scrolling region */
  BOOL     NsrClosed;                   /* TRUE when </div> for closing NSR sent */
  int      rtf_sl;                      /* RTF-Parameter Line Spacing */
  int      rtf_sa;                      /* RTF-Parameter Spacing After */
  int      rtf_sb;                      /* RTF-Parameter Spacing Before */
  int      rtf_li;                      /* RTF-Parameter Left Indent */
  int      rtf_ri;                      /* RTF-Parameter Right Intent */
  int      rtf_fi;                      /* RTF-Parameter First Intent */
  int      BgColorNsr;                  /* Background-Color in Non-Scrolling Region */

  BOOL     FormatNavigation;            /* TRUE wenn Formatting with navigation elements (contents, prev, next...) */
  BOOL     FormatPopup;                 /* TRUE when displaying in a popup */
  BOOL     FormatPrint;                 /* TRUE when Format for Printing (window.print()) */
  BOOL     FormatMarkPopup;             /* TRUE if popup-Jumps shall be specially marked as such */
  BOOL     FormatAddHlpTitle;           /* TRUE to add the title of the .hlp file to the title of the topic .html file */

  const char *virtual_root;             /* Directory name of the virtual root of the help file, for example "../.." */

  HTOPIC   topic;                       /* Topic file handle obtained from HlpTopicOpen */
  LPCSTR   HlpContextId;                /* Contents ID of current topic */

  char *   keyword;                     /* Keywords encountered in this topic */

  TABSETTINGS tabs_rtf;                 /* Currently built tab settings */
  int      tabalign;                    /* Alignment for the next tabstop that will be defined */
  int      tabcurrent;                  /* Current Tab position */
  TABSETTINGS tabs_html;                /* Tabs currently handled on HTML side */

  BOOL        InTabEmulationTable;      /* TRUE as long as the \tab emulation table is open */
  BOOL        InAlignedBitmap;          /* True of an aligned bitmap has been set */

  STRINGBUFFER sbrow;                   /* Buffer to hold a table row before advanding. */
  STRINGBUFFER sbtd;                    /* Buffer for last <td...> in tab emulation */
  STRINGBUFFER *sbout;                  /* Output STRINGBUFFER that will old the conversion result */

  IMAGEINFO imageinfo[MAX_IMAGES];      /* Informations cached on images */
} ENUMERATION_PRIVATE_DATA, FAR *LPENUMERATION_PRIVATE_DATA;


/* ------ Local Function prototypes ---------------------------------- */
static  void SimpleRtfToHtml(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszRtf);
static  void ResetFont(LPENUMERATION_PRIVATE_DATA dta);
static  void _OpenPara(LPENUMERATION_PRIVATE_DATA dta);
static  void _CloseHtmlPara(LPENUMERATION_PRIVATE_DATA dta);
static  BOOL _IsEmptyString(const char *str);
static  void _OpenTabEmulationTable(LPENUMERATION_PRIVATE_DATA dta);

/* _CreateAnchor
 *
 * Creates an anchor referencing to the Specified information.
 */
static
void
_CreateAnchor(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszTopicText)
{
  char  HlpFileName[MAX_PATH],
        WindowName[50],
        ContextId[200];
  LPSTR dst=ContextId;
  HIFS  NewIFS;
  DWORD ContextIdTopicOffset,
        TitleTopicOffset;

  WindowName[0]=0;
  HlpFileName[0]=0;

  _OpenPara(dta);

  /* Seperate Jump destination into Context-Id, Window-Name and Filename */
  strcpy(HlpFileName, dta->HlpFileName);

  while (*lpszTopicText)
  {
    switch (*lpszTopicText)
    {
       case '@': *dst=0; dst=HlpFileName; break;
       case '>': *dst=0; dst=WindowName; break;
       default : *dst++=*lpszTopicText;
    }
    lpszTopicText++;
  }
  *dst=0;

  /* Look if jump-Destination is currently available */
  NewIFS=IFSOpen(HlpFileName);
  if (NewIFS)
  {
     if (0xffffffffL != (ContextIdTopicOffset = HlpIsTopic(NewIFS, HlpGenerateHashValue(ContextId))) &&   // Context If can be found.
         HlpGetTopicTitle(NewIFS, ContextIdTopicOffset, NULL, 0, &TitleTopicOffset))                      // Need context-Id for start of Topic
     {
       char virtname[500];

       ResetFont(dta);
       
       if (ContextIdTopicOffset != TitleTopicOffset)
         {
           /* This is a Jump to a sub-anchor of the topic. */
           LPSTR TitleContextID = HlpTopicFindContextIdForTopicOffset(dta->topic, TitleTopicOffset);
           if (TitleContextID)
             {
               snprintf(virtname, sizeof(virtname),
                        "id/%s.html#%s",
                        UrlEscape(TitleContextID),
                        UrlEscape(ContextId));
             }
           else
             {
               snprintf(virtname, sizeof(virtname),
                        "topicoffset/0x%08x.html#%s",
                        TitleTopicOffset,
                        UrlEscape(ContextId));
               
             }
         }
       else
         {
           snprintf(virtname, sizeof(virtname),
                    "id/%s.html",
                    UrlEscape(ContextId));
         }


       /* Setup HTML anchor */
       if (WindowName[0]!=0 && strcmp(WindowName, "main") && dta->FormatMarkPopup)
         {
           sbprintf(&dta->sbrow, "<a href=\"%s/popup/%s\" target=\"%s\" onclick=\'window.open(\"%s/popup/%s\", \"%s\", \"width=400,height=400,screenX=0,screenY=0,dependent=yes,locationbar=no,menubar=no,status=no\")'>",
                  dta->virtual_root, virtname, WindowName, 
                  dta->virtual_root, virtname, WindowName);
         }
       else if (dta->FormatPopup)
         {
           sbprintf(&dta->sbrow, "<a href=\"%s/%s\" target=\"main\" onclick=\'opener.location=\"%s/%s\"; return false;'>", 
                  dta->virtual_root, virtname, 
                  dta->virtual_root, virtname);
         }
       else
         {
           sbprintf(&dta->sbrow, "<a href=\"%s/%s\">", dta->virtual_root, virtname);
         }
        dta->InAnchor=TRUE;
     }

     IFSClose(NewIFS);
  }
}


/**
 * Process a macro-string to create an anchor from it if 
 * It points to some Internet-Location, for example an EF("http://www.herdsoft.com/") macro
 *
 * @param dta             Enumeration object
 * @param lpszMacroString The Macro String.
 */
static
void
_ProcessMacroString(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszTopicText)
{
  const char *p;

  if (NULL!=(p=strstr(lpszTopicText,"http://")) ||
      NULL!=(p=strstr(lpszTopicText,"https://")) ||
      NULL!=(p=strstr(lpszTopicText,"mailto:")) )
    {
      char url[1000];
      char *dest;
      for (dest=url; *p && !strchr("\"'", *p) && (dest-url)<sizeof(url)-1; )
        *dest++=*p++;
      *dest=0;

      ResetFont(dta);
      _OpenPara(dta);
      sbprintf(&dta->sbrow, "<a href=\"%s\">", url);
      dta->InAnchor=TRUE;
    }
}





/* ResetFont
 *
 * Reset font settings.
 */
static
void
ResetFont(LPENUMERATION_PRIVATE_DATA dta)
{
  if (dta->FontSet)
	 {
		if (dta->lflast.lfUnderline)                                     sbprintf(&dta->sbrow, "</u>");
		if (dta->lflast.lfItalic)                                        sbprintf(&dta->sbrow, "</i>");
		if (dta->lflast.lfWeight>FW_NORMAL)                              sbprintf(&dta->sbrow, "</b>");
		
		sbadd(&dta->sbrow, "</font>");
		
		dta->FontSet = FALSE;
	 }
} // ResetFont


/* SetFont
 *
 * Set the font according to the selected parameters 
 */
void
SetFont(LPENUMERATION_PRIVATE_DATA dta)
{
  ResetFont(dta);

  sbprintf(&dta->sbrow, "<font size=\"%d\" face=\"%s\"",
			dta->lfnow.lfHeight/5,
			dta->lfnow.lfFaceName
			);

  if (!dta->InAnchor && (dta->clnow.rgbRed>1||dta->clnow.rgbGreen>1||dta->clnow.rgbBlue>1))
    {
      sbprintf(&dta->sbrow, " color=\"#%02x%02x%02x\"",
             dta->clnow.rgbRed, dta->clnow.rgbGreen, dta->clnow.rgbBlue);
    }

  sbputchar(&dta->sbrow, '>');

  if (dta->lfnow.lfWeight>FW_NORMAL)                               sbprintf(&dta->sbrow, "<b>");
  if (dta->lfnow.lfItalic)                                         sbprintf(&dta->sbrow, "<i>");
  if (dta->lfnow.lfUnderline)                                      sbprintf(&dta->sbrow, "<u>");

  dta->cllast = dta->clnow;
  dta->lflast = dta->lfnow;

  dta->FontSet = TRUE;
  //  memcpy(&dta->cllast, &dta->clnow, sizeof(RGBQUAD));
  //memcpy(&dta->lflast, &dta->lfnow, sizeof(LOGFONT));
} // SetFont


/**
 * Process an RTF \tab command. Advance to the next tab position.
 * Also called before the first tab when opening the paragraph
 */
static
void
_NextTab(LPENUMERATION_PRIVATE_DATA dta)
{
  const char *align;
  TABSTOP *tstop;

  _CloseHtmlPara(dta);

  if (dta->tabcurrent != 0 &&
      dta->tabcurrent <= dta->tabs_html.tabcount)
    {
      sbadd(&dta->sbrow, "\n </td>");
    }

  /* Flush what we had so far... */
  sbcat(dta->sbout, &dta->sbtd);
  sbcat(dta->sbout, &dta->sbrow);

  if (dta->tabcurrent < dta->tabs_html.tabcount)
    {
      int xleft;

      tstop = dta->tabs_html.tabstop + dta->tabcurrent;
      
      switch (tstop->tabalign)
        {
        case 1: align="center"; break;
        case 2: align="right"; break;
        default: align="left"; break;
        }
      
      xleft=dta->tabcurrent==0 ? 0 : tstop[-1].tabpos;
      /* Remeber td command to add on next \tab */
/*       if (dta->tabcurrent < dta->tabs_html.tabcount) */
/*         { */
/*           sbprintf(&dta->sbtd, "\n <td valign=\"top\" align=\"%s\" width=\"%d\">\n", align, */
/*             (tstop->tabpos - xleft) * 180 / 1440); */
/*         } */
/*       else */
        {
          sbprintf(&dta->sbtd, "\n <td valign=\"top\" align=\"%s\">\n", align);
        }
    }
  else if (dta->tabcurrent == dta->tabs_html.tabcount && dta->tabcurrent != 0)
    {
      /* Remeber td command to add on next \tab */
      sbprintf(&dta->sbtd, "\n <td valign=\"top\" align=\"left\">\n");
    }

  if (dta->tabs_html.tabcount != 0)
    {
      dta->tabcurrent++;
    }
}


/**
 * Process a RTF \line command to advance to the next line in the same paragraph
 *
 * Normally this is simply a <br>, but if we are emulating tabstops then it is much more
 * complex.
 */
static
void
_RtfLineCommand(LPENUMERATION_PRIVATE_DATA dta)
{
  if (dta->tabcurrent==0)
    {
      /* Simple mode. */
      sbadd(&dta->sbrow, "<br>\n");
    }
  else
    {
      _CloseHtmlPara(dta);

      sbcat(dta->sbout, &dta->sbtd);
      sbcat(dta->sbout, &dta->sbrow);

      /* Tab emulation */
      sbadd(&dta->sbrow,
             "\n"
             " </td>\n"
             "</tr>\n"
             "<tr>");

      dta->tabcurrent=0;

      _NextTab(dta);
      
      /* Advance table columns until we are in the right position */
      while (dta->tabcurrent<=dta->tabs_html.tabcount &&
             dta->tabs_html.tabstop[dta->tabcurrent-1].tabpos <= dta->rtf_li)
        {
          _NextTab(dta);
        }
    }
}


/**
 * Close the tab-emulation table if it is still open.
 */
static
void
_CloseTabEmulationTable(LPENUMERATION_PRIVATE_DATA dta)
{
  if (dta->InTabEmulationTable)
    {
      sbadd(&dta->sbrow, "</table>\n");
      dta->InTabEmulationTable = FALSE;
      dta->tabcurrent = 0;
      dta->tabs_html.tabcount = 0;
    }
}

/**
 * Open the tab emulation table if it is going to be used for the next RTF-Paragraph+
 */
static
void
_OpenTabEmulationTable(LPENUMERATION_PRIVATE_DATA dta)
{
  if (!dta->InNsr && !dta->NsrClosed)
    {
      dta->NsrClosed = TRUE;
      
      _CloseTabEmulationTable(dta);
      sbadd(&dta->sbrow, "\n</div>\n");
    }
  
  if (dta->tabs_rtf.tabcount)
    {
      if (dta->tabcurrent==0)
        {
          /* We want tab emulation. */
          if (dta->InTabEmulationTable)
            {
              /* A tab emulation table has already been opened.
               * Leave it open if tab settings have not been changed.
               */
              if (memcmp(&dta->tabs_html, &dta->tabs_rtf, sizeof(dta->tabs_html)))
                {
                  /* Tab parameters have been modified. Close table and open new one later. */
                  _CloseTabEmulationTable(dta);
                }
            }
          
          if (!dta->InTabEmulationTable)
            {
              memcpy(&dta->tabs_html, &dta->tabs_rtf, sizeof(dta->tabs_html));
              sbadd(&dta->sbrow, "\n<table border=\"0\">\n");
              dta->InTabEmulationTable = TRUE;
            }
          
          sbadd(&dta->sbrow, "<tr>\n");
          _NextTab(dta);
        }
    }
  else
    {
      /* No tab emulation needed for this paragraph */
      _CloseTabEmulationTable(dta);
    }
}


/**
 * Open a Paragraph (if not already done) and set Paragraph style parameters
 */
static
void
_OpenPara(LPENUMERATION_PRIVATE_DATA dta)
{
  if (!dta->InPara)
    {
      _OpenTabEmulationTable(dta);

      sbadd(&dta->sbrow, "<p style=\"");
      if (dta->rtf_sa != 0)
        {
          sbprintf(&dta->sbrow, "margin-bottom:%.1fpt;", dta->rtf_sa / 20.0);
        }
      if (dta->rtf_sl != 0)
        {
          sbprintf(&dta->sbrow, "line-height:%.1fpt;", dta->rtf_sl / 20.0);
        }
      if (dta->rtf_sb != 0)
        {
          sbprintf(&dta->sbrow, "margin-top:%.1fpt;", dta->rtf_sb / 20.0);
        }

      if (dta->tabs_html.tabcount==0)
        {
          if (dta->rtf_li != 0)
            {
              sbprintf(&dta->sbrow, "margin-left:%.1fpt;", dta->rtf_li / 20.0);
            }
          if (dta->rtf_ri != 0)
            {
              sbprintf(&dta->sbrow, "margin-right:%.1fpt;", dta->rtf_ri / 20.0);
            }
          if (dta->rtf_fi != 0)
            {
              sbprintf(&dta->sbrow, "text-indent:%.1fpt;", dta->rtf_fi / 20.0);
            }
        }
      
      if (dta->InNsr && dta->BgColorNsr!=0xffffff)
        {
          sbprintf(&dta->sbrow, "background-color: #%6x;", dta->BgColorNsr);
        }

      if (dta->InBrdrl)
        {
          sbadd(&dta->sbrow, "border-left-width:thin; border-left-style:solid;");
        }
      if (dta->InBrdrr)
        {
          sbadd(&dta->sbrow, "border-right-width:thin; border-right-style:solid;");
        }
      if (dta->InBrdrt)
        {
          sbadd(&dta->sbrow, "border-top-width:thin; border-top-style:solid;");
        }
      if (dta->InBrdrb)
        {
          sbadd(&dta->sbrow, "border-bottom-width:thin; border-bottom-style:solid;");
        }

      sbadd(&dta->sbrow, "\">");
      
      dta->InPara=TRUE;
    }
}


/**
 * Close a paragraph. HTML side.
 *
 * One paragraph on the RTF side might contain several paragraphs on the
 * HTML side, one of each tabstop, each one packed in a <td><p>...</p></td>
 * This function closes the HTML side.
 */
static
void
_CloseHtmlPara(LPENUMERATION_PRIVATE_DATA dta)
{
  if (dta->InPara)
    {
      dta->InPara = FALSE;
      
      ResetFont(dta);

      if (dta->InAlignedBitmap)
        {
          /* In HTML aligned Bitmaps span paragraph boundaries. 
           * In RTF they don't. This causes strange side effects.
           * So cancel the bitmap alignment here.
           */
          dta->InAlignedBitmap = FALSE;
          sbadd(&dta->sbrow, "<br clear=\"all\">");
        }

      sbadd(&dta->sbrow, "</p>\n");
    }
}

/**
 * Close a paragraph.RTF side.
 *
 * One paragraph on the RTF side might contain several paragraphs on the
 * HTML side, one of each tabstop, each one packed in a <td><p>...</p></td>
 *
 * This function closes the RTF side.
 */
static
void
_CloseRtfPara(LPENUMERATION_PRIVATE_DATA dta)
{
  _CloseHtmlPara(dta);

  if (dta->tabcurrent != 0)
    {
      int colspan = dta->tabs_html.tabcount - dta->tabcurrent + 2;
      if (colspan > 1)
        {
          /* The paragraph defined more tabstops than it included tabs.
           * This is a commen problem due to a bug in the windows help compiler
           * (at least in some versions)
           *
           * Use a <td colspan=...> formatting to avoid the browser to
           * misdisplay this paragraph.
           */
          sbclear(&dta->sbtd);
          sbprintf(&dta->sbtd, 
                    "<td valign=\"top\" align=\"left\" colspan=\"%d\">\n",
                    colspan);
        }

      dta->tabcurrent = 0;

      sbcat(dta->sbout, &dta->sbtd);
      sbcat(dta->sbout, &dta->sbrow);

      sbadd(&dta->sbrow, " </td>\n</tr>\n");
    }
}




/**
 * Process an image reference to produce an image map for the image.
 *
 * @param dta        Enumeration data Object
 * @param imagename  Image Name, for example "bm1"
 * @param pwidth     On return contains width of the image for <img width=...>
 * @param pheight    On return contains height of the image for <img width=...>
 *
 * @return TRUE if the file processed contained an image map.
 */
static
BOOL
_ProcessImageMap(LPENUMERATION_PRIVATE_DATA dta, const char *imagename, int *pwidth, int *pheight)
{
  char fullimagename[50];
  HIFSFILE  ifsfile;
  BOOL result = FALSE;

  snprintf(fullimagename, sizeof(fullimagename),
           "|%s",
           imagename);

  if (NULL!=(ifsfile=IFSOpenFile(dta->ifs, fullimagename)))
    {
      DWORD     dwSHGSize;
      LPSTR     lpSHG;
      
      /* Get file size */
      dwSHGSize = IFSSeekFile(ifsfile, 0, SEEK_END);
      IFSSeekFile(ifsfile, 0, SEEK_SET);

      /* Read whole image into memory */
      lpSHG = (LPSTR) malloc(dwSHGSize);
      if (lpSHG)
        {
          METAFILEHEADER metaFileHeader;
          HGLOBAL hseg=NULL;

			 IFSReadFile(ifsfile, lpSHG, dwSHGSize);
          memset(&metaFileHeader, 0, sizeof(metaFileHeader));
          
          /* Convert SHG-File to standard DIB or Metafile format */
			 SHGImport(
#ifdef _Windows
                    NULL,                           // Parent-Window for error Messages
#endif
						  lpSHG,                          // Input Data Block
						  NULL,                           // DIB-Handle
						  NULL,                           // Metafile-Handle for SHG containing WMF
						  NULL,                           // Will be filled with Informations on placeable Metafiles
                    &metaFileHeader,                // Placeable Metafile header Informations
						  &hseg                           // Segmentation Informations
						  );
			 
          if (hseg)
            {
              SEGINFOHEADER *pseg = (SEGINFOHEADER *) GlobalLock(hseg);
              int i;

              if (metaFileHeader.inch)
                {
                  *pwidth = (metaFileHeader.bbox.right  - metaFileHeader.bbox.left) * 96 / metaFileHeader.inch;
                  *pheight = (metaFileHeader.bbox.bottom  - metaFileHeader.bbox.top) * 96 / metaFileHeader.inch;
                }
              else
                {
                  *pwidth = pseg->Width;
                  *pheight = pseg->Height;
                }

              if (pseg->SegJumpCount > 0)
                {
                  SEGJUMP *pjump = (SEGJUMP *) (pseg+1);

                  sbprintf(&dta->sbrow, "<map name=\"%s\">", imagename);
                  for (i=0; i<pseg->SegJumpCount; i++, pjump++)
                    {
                      if ( !(pjump->flags & SI_MACRO) )
                        {
                          int left, top, right, bottom;

                          const char *macro = ( (const char *) pseg) + pjump->macro.dw;

                          if (metaFileHeader.inch)
                            {
                              left = (pjump->left - metaFileHeader.bbox.left) * 96 / (metaFileHeader.inch*10);
                              top  = (pjump->top - metaFileHeader.bbox.top) * 96 / (metaFileHeader.inch*10);
                              right = (pjump->right - metaFileHeader.bbox.left) * 96 / (metaFileHeader.inch*10);
                              bottom  = (pjump->bottom - metaFileHeader.bbox.top) * 96 / (metaFileHeader.inch*10);
                            }
                          else
                            {
                              left = pjump->left;
                              top  = pjump->top;
                              right = pjump->right;
                              bottom  = pjump->bottom;
                            }

                          sbprintf(&dta->sbrow, 
                                   "<area shape=\"rect\" coords=\"%d,%d,%d,%d\" href=\"%s/%s/%s.html\">\n",
                                   left, top, right, bottom,
                                   dta->virtual_root, 
                                   pjump->flags & SI_POPUP ? "popup/id" : "id",
                                   UrlEscape(macro));

                          result = TRUE;
                        }
                    }
                  sbprintf(&dta->sbrow, "</map>\n");
                }

              GlobalUnlock(hseg);
              GlobalFree(hseg);
            }
          
          free(lpSHG);
        }

      IFSCloseFile(ifsfile);
    }

  return result;
}


/* CreateImageRef
 *
 * Creates an reference to an image
 */
void CreateImageRef(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszTopicText)
{
  LPCSTR p=strchr(lpszTopicText, ' ');
  LPCSTR szAlign="";
  int imagenum=-1;

  int width=-1;
  int height=-1;
  BOOL usemap;

  switch (*(p-1))
  {
    case 'l' : szAlign=" align=left"; dta->InAlignedBitmap=TRUE; break;
    case 'r' : szAlign=" align=right"; dta->InAlignedBitmap=TRUE; break;
  }

  p++;
  *(strchr(p,'}'))=0;

  _OpenPara(dta);

  if (*p=='|')
    p++;


  if (p[0]=='b'&&p[1]=='m')
    {
      imagenum = atoi(p+2);
    }

  if (imagenum>=0 && imagenum<MAX_IMAGES &&
      dta->imageinfo[imagenum].tested)
    {
      usemap = dta->imageinfo[imagenum].usemap;
      width = dta->imageinfo[imagenum].width;
      height = dta->imageinfo[imagenum].height;
    }
  else
    {
      usemap = _ProcessImageMap(dta, p, &width, &height);

      if (imagenum>=0 && imagenum<MAX_IMAGES)
        {
          dta->imageinfo[imagenum].width = width;
          dta->imageinfo[imagenum].height = height;
          dta->imageinfo[imagenum].usemap = usemap;
          dta->imageinfo[imagenum].tested = TRUE;
        }
    }

  sbprintf(&dta->sbrow, "<img src=\"%s/png/%s.png\"%s alt=\"\" border=\"0\"", dta->virtual_root, p, szAlign);

  if (usemap)
    {
      sbprintf(&dta->sbrow, " usemap=\"#%s\"", p);
    }

  if (width>0 && height>0)
    {
      sbprintf(&dta->sbrow, " width=\"%d\" height=\"%d\"", width, height);
    }

  sbputchar(&dta->sbrow, '>');
} // CreateImageRef


/* TopicEnumerationLinkRelProcedure
 *
 * This is the callback-Function this application defined for the call to
 * HlpTopicEnumerate to Enumerate the TOPIC-Files contents
 *
 * This version is just responsible to create the <link rel=...> tags in the header
 */
BOOL 
FAR PASCAL 
TopicEnumerationLinkRelProcedure(
                                 LPCSTR    lpszTopicText,   // Text to be transfered to the Application program
                                 DWORD     dummy,           // resevred
                                 TEXTTYPE  eTextType,       // type of Text-String transfered to the Application
                                 LPARAM    lParam)          // Application-Defined user Data
{
  BOOL result = TRUE;
  LPENUMERATION_PRIVATE_DATA dta = (LPENUMERATION_PRIVATE_DATA) lParam;
  
  switch (eTextType)
	 {
	 case TOPIC_BROWSE_BACK:
	 case TOPIC_BROWSE_FORWARD:
		{
		  DWORD dwTopicOffset;
		  char  szTopicTitle[512];
		  sscanf(lpszTopicText, "%8x", &dwTopicOffset);
		  
		  if (dwTopicOffset!=0xffffffffL && dta->FormatNavigation)
			 {
            LPCSTR contextID;
				*szTopicTitle=0;
				HlpGetTopicTitle(dta->ifs, dwTopicOffset, szTopicTitle, sizeof(szTopicTitle)-1, &dwTopicOffset);
				
            contextID = HlpTopicFindContextIdForTopicOffset(dta->topic, dwTopicOffset);
            if (contextID)
              {
                sbprintf(&dta->sbrow, "<link rel=\"%s\" href=\"%s/id/%s.html\" title=\"",
                       eTextType==TOPIC_BROWSE_FORWARD ? "next" : "prev",
                       dta->virtual_root,
                       contextID);
              }
            else
              {
                sbprintf(&dta->sbrow, "<link rel=\"%s\" href=\"%s/topicoffset/0x%s.html\" title=\"",
                       eTextType==TOPIC_BROWSE_FORWARD ? "next" : "prev",
                       dta->virtual_root,
                       lpszTopicText);
              }

				sbhtmlputs(&dta->sbrow, szTopicTitle);
            sbadd(&dta->sbrow, "\">\n");
			 }

        if (eTextType==TOPIC_BROWSE_BACK)
          {
            if (!dta->BrowseBackCalled)
              {
                dta->BrowseBackCalled=TRUE;
          }
            else
              {
                result = FALSE;
              }
          }
        break;

		}
		break;
   
/* 	 case TOPIC_TEXT:                   // Normal Text... */
/*       result = FALSE; */
/*       break; */

    case TOPIC_KEYWORD:
      if (lpszTopicText[0]=='K' &&
          lpszTopicText[1]=='\t')
        {
          lpszTopicText += 2;
        }

      if (dta->keyword)
        {
          int lenn=strlen(lpszTopicText);
          int leno=strlen(dta->keyword);
          dta->keyword = (char *) realloc(dta->keyword, lenn + leno + 2);
          sprintf(dta->keyword+leno, ",%s", lpszTopicText);
        }
      else
        {
          dta->keyword = strdup(lpszTopicText);
        }
      break;
      
    default: ;
    }

  return result;
}


/* TopicEnumerationProcedure
 *
 * This is the callback-Function this application defined for the call to
 * HlpTopicEnumerate to Enumerate the TOPIC-Files contents
 */
BOOL 
FAR PASCAL 
TopicEnumerationProcedure(
								  LPCSTR    lpszTopicText,   // Text to be transfered to the Application program
								  DWORD     dummy,           // resevred
								  TEXTTYPE  eTextType,       // type of Text-String transfered to the Application
								  LPARAM    lParam)          // Application-Defined user Data
{
  LPENUMERATION_PRIVATE_DATA dta = (LPENUMERATION_PRIVATE_DATA) lParam;
  
  switch (eTextType)
	 {
	 case TOPIC_RTF_INIT:                     // RTF Control strings
	 case TOPIC_RTF_TEXT:                     // RTF Control strings
	 case TOPIC_RTF_EXIT:                     // RTF Control strings
		SimpleRtfToHtml(dta, lpszTopicText);
		break;
		
	 case TOPIC_TEXT:                   // Normal Text...
		
		if (dta->HtmlTable && !dta->RtfTable)
		  {  
			 /* Theoretically a Table always spans only one line in RTF, so we have to collect them */
			 dta->HtmlTable=FALSE;
			 
			 /* Always close font when changing table items */
			 _CloseHtmlPara(dta);

			 if (dta->HtmlRow)
				{
				  sbadd(&dta->sbrow, "</td></tr>\n");
				}
			 dta->HtmlRow=FALSE;
			 sbadd(&dta->sbrow, "</table>\n");
		  }
		
		if (dta->HtmlTable && !dta->HtmlRow)
		  {
			 /* Always close font when changing table items */
			 _CloseHtmlPara(dta);

			 sbadd(&dta->sbrow, "<tr><td valign=\"top\">\n");
			 dta->HtmlRow=TRUE;
		  }
		
      if (!_IsEmptyString(lpszTopicText) )
        {
          _OpenPara(dta);
        }

		if (memcmp(&dta->clnow,&dta->cllast,sizeof(RGBQUAD))||
			 memcmp(&dta->lfnow,&dta->lflast,sizeof(LOGFONT))||
			 !dta->FontSet)
		  {
			 if (*lpszTopicText)
				{
				  SetFont(dta);
				}
		  }
		
		sbhtmlputs(&dta->sbrow, lpszTopicText);
		break;
		
	 case TOPIC_BROWSE_BACK:
	 case TOPIC_BROWSE_FORWARD:
		{
		  DWORD dwTopicOffset;
		  char  szTopicTitle[512];
        const char *makeBorder = " style=\"width:80px; text-align:center; border-width:medium; border-color:#808080; background-color: #c0c0c0; border-style:outset\"";

        const char *http_user_agent;
        if (dta->FormatNavigation &&
            NULL!=(http_user_agent = getenv("HTTP_USER_AGENT")) &&
            NULL!=strstr(http_user_agent, "Mozilla/4.") &&
            NULL==strstr(http_user_agent, "MSIE "))
          {
            makeBorder = "";
          }

        if (eTextType==TOPIC_BROWSE_BACK)
          {
            if (!dta->BrowseBackCalled)
              {
                dta->BrowseBackCalled=TRUE;
              }
            else
              {
                return FALSE;
              }
          }

		  if (dta->FormatNavigation)
          {
            if (eTextType==TOPIC_BROWSE_BACK)
              {
                const char *script_name = getenv("SCRIPT_NAME");
                if (script_name != NULL)
                  {
                    sbprintf(&dta->sbrow, "<p>\n<a href=\"%s\"%s title=\"%s\">%s</a>\n",
                             script_name,
                             makeBorder,
                             _("List of available help files"),
                             _("Directory"));
                  }
                sbprintf(&dta->sbrow, "<a href=\"%s/\"%s title=\"%s\">%s</a>\n",
                         dta->virtual_root,
                         makeBorder,
                         _("Display the contents listing of this Help File"),
                         _("Contents"));
                sbprintf(&dta->sbrow, "<a href=\"%s/?KEYWORD=\"%s title=\"%s\">%s</a>\n",
                       dta->virtual_root,
                       makeBorder,
                       _("Search the current help file for keywords"),
                       _("Search")
                       );
                if (dta->HlpContextId)
                  {
                    sbprintf(&dta->sbrow, "<a href=\"%s/print/id/%s.html\"%s title=\"%s\" target=\"print\">%s</a>\n",
                           dta->virtual_root,
                           UrlEscape(dta->HlpContextId),
                           makeBorder,
                           _("Print the current topic"),
                           _("Print")
                           );
                  }
                else
                  {
                    sbprintf(&dta->sbrow, "<a href=\"%s/print/topicoffset/0x%08x.html\"%s title=\"%s\" target=\"print\">%s</a>\n",
                           dta->virtual_root,
                           dta->TopicOffset,
                           makeBorder,
                           _("Print the current topic"),
                           _("Print")
                           );
                  }
              }
            
            sscanf(lpszTopicText, "%8x", &dwTopicOffset);
            
            if (dwTopicOffset!=0xffffffffL)
              {
                LPCSTR contextID;

                *szTopicTitle=0;
                HlpGetTopicTitle(dta->ifs, dwTopicOffset, szTopicTitle, sizeof(szTopicTitle)-1, &dwTopicOffset);
                
                contextID = HlpTopicFindContextIdForTopicOffset(dta->topic, dwTopicOffset);
                if (contextID)
                  {
                    sbprintf(&dta->sbrow, "<a href=\"%s/id/%s.html\"%s title=\"",
                           dta->virtual_root,
                           contextID,
                           makeBorder);
                  }
                else
                  {
                    sbprintf(&dta->sbrow, "<a href=\"%s/topicoffset/0x%s.html\"%s title=\"",
                           dta->virtual_root,
                           lpszTopicText,
                           makeBorder);
                  }
                sbhtmlputs(&dta->sbrow, szTopicTitle);
                sbprintf(&dta->sbrow, "\">%s</a>\n",
                       eTextType==TOPIC_BROWSE_FORWARD ? "&gt;&gt;" : "&lt;&lt;");
              }
            
            if (eTextType==TOPIC_BROWSE_FORWARD)
              {
                sbprintf(&dta->sbrow, 
                         "</p>"
                         "<hr>\n");
              }
          }
		}
		break;
		
	 case TOPIC_HOTSPOT_START:          // reference to another topic or Anchor
		if (*lpszTopicText=='%' || // Invisible Hotspots...
			 *lpszTopicText=='*')
		  lpszTopicText++;

      _CreateAnchor(dta, lpszTopicText);
		break;

    case TOPIC_MACRO_HOTSPOT:
		if (*lpszTopicText=='%' || // Invisible Hotspots...
			 *lpszTopicText=='*')
		  lpszTopicText++;

      _ProcessMacroString(dta, lpszTopicText);
        
      break;
		
	 case TOPIC_HOTSPOT_END:            // reference to another topic or Anchor END
		if (dta->InAnchor)
        {
          ResetFont(dta);
          
          sbadd(&dta->sbrow, "</a>");
        }
		dta->InAnchor=FALSE;
		break;
		
	 case TOPIC_BMX:                    // Reference to a Bitmap (or Metafile :-( )
		CreateImageRef(dta, lpszTopicText);
      break;

      
    case TOPIC_CONTEXTID:
      /// Found a context-ID (# footnote)
      sbprintf(&dta->sbrow, "\n\n<a name=\"%s\"></a>\n", UrlEscape(lpszTopicText));
      break;

    default: ;
	 }
  
  return TRUE;                          // Continue enumeration
}




/* EnumerateTopic
 *
 * Enumerates the Contents of one Topic of a Help File.
 *
 * @param sbout             Output STRINGBUFFER to receive the resulting HTML content
 * @param ifs               Handle to the internal file system created by IFSOpen
 * @param topic             Handle for |TOPIC file created by HlpTopicOpen
 * @param HlpFileName       Filename of the .hlp or .mvb file.
 * @param HlpTopicOffset    Topic-Offset stored as a string
 * @param HlpContextId      Context-ID of the topic that shall be displayed
 * @param dwFormattingFlags Flags to influence formatting of the resulting HTML
 * @param virtual_root      URL Reference to the virtual root of the help file.
 */
void EnumerateTopic(STRINGBUFFER *sbout, HIFS ifs, HTOPIC topic, LPCSTR HlpFileName, LPCSTR HlpTopicOffset, LPCSTR HlpContextId, DWORD dwFormattingFlags, LPCSTR virtual_root)
{ 
  DWORD                    TopicOffset=0;
  LPENUMERATION_PRIVATE_DATA dta=(LPENUMERATION_PRIVATE_DATA) calloc(sizeof(ENUMERATION_PRIVATE_DATA), 1);
  char                     szTopicTitle[512];
  int                      BgColor=0xffffff;

  sbadd(sbout, 
        "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
        "<html>\n"
        "<head>\n");

  if (dta)
	 {
      dta->topic = topic;
      dta->sbout = sbout;

      if (HlpTopicOffset!=NULL)
        {
          sscanf(HlpTopicOffset+2, "%x", &TopicOffset);
        }
      else if (HlpContextId!=NULL)
        {
          TopicOffset = HlpIsTopic(ifs, HlpGenerateHashValue(HlpContextId));
        }
		
      /* ------------- Inclue Topic Title as HTML-Title ------------------*/
      sbprintf(&dta->sbrow, "<title>");

      {
        LPHELPFILETITLE lpHelpTitle;
        lpHelpTitle = HlpGetHelpTitle(ifs);
        if (lpHelpTitle )
          {
            if (lpHelpTitle->HelpTitle && (dwFormattingFlags & HTMLTOPIC_ADDHLPTITLE)!=0)
              {
                sbhtmlputs(&dta->sbrow, lpHelpTitle->HelpTitle);
                sbprintf(&dta->sbrow, " - ");
              }
            BgColor = lpHelpTitle->BgColor;
            dta->BgColorNsr = lpHelpTitle->BgColorNsr;
            HlpFreeHelpTitle(lpHelpTitle);
          }
      }


      HlpGetTopicTitle(ifs, TopicOffset, szTopicTitle, sizeof(szTopicTitle), &TopicOffset);
      sbhtmlputs(&dta->sbrow, szTopicTitle[0] ? szTopicTitle : _("<No Title - maybe popup...>"));
      sbadd(&dta->sbrow, "</title>\n");

      dta->topic           = topic;
      dta->BrowseBackCalled= FALSE;
      dta->HlpFileName     = HlpFileName;
      dta->TopicOffset     = TopicOffset;
      dta->HlpContextId    = HlpContextId;
      dta->ifs             = ifs;
      dta->FormatNavigation= dwFormattingFlags & HTMLTOPIC_NAVIGATION;
      dta->FormatPopup     = dwFormattingFlags & HTMLTOPIC_POPUP;
      dta->FormatPrint     = dwFormattingFlags & HTMLTOPIC_PRINT;
      dta->FormatMarkPopup = dwFormattingFlags & HTMLTOPIC_MARKPOPUP;
      dta->FormatAddHlpTitle= dwFormattingFlags & HTMLTOPIC_ADDHLPTITLE;
      dta->virtual_root    = virtual_root;
      
      if (dta->FormatPopup || dta->FormatPrint)
        {
          sbprintf(&dta->sbrow, "<meta name=\"robots\" content=\"noindex\">\n");
        }
      
      sbadd(&dta->sbrow, "<meta name=\"creator\" content=\"winhelpcgi " VERSION "\">\n");
      if (dta->FormatNavigation)
        {
          sbprintf(&dta->sbrow, "<link rel=\"up\" href=\"%s/\" title=\"Contents\">\n", dta->virtual_root);
          sbprintf(&dta->sbrow, "<link rel=\"start\" href=\"%s/../\" title=\"Start\">\n", dta->virtual_root);
          sbprintf(&dta->sbrow, "<link rel=\"bookmark\" href=\"http://www.herdsoft.com/\" title=\"Herd Software Development\">\n");
          sbprintf(&dta->sbrow, "<link rel=\"bookmark\" href=\"http://www.herdsoft.com/catalog/ehlp2rtf.html\" title=\"Help to RTF Converter\">\n");
          sbprintf(&dta->sbrow, "<link rel=\"bookmark\" href=\"http://www.herdsoft.com/linux/produkte/winhelpcgi.html\" title=\"WinHelpCGI project page\">\n");
          sbprintf(&dta->sbrow, "<link rel=\"search\" href=\"%s/?KEYWORD=\" title=\"Glossary\">\n", dta->virtual_root);
        }
          
      //------------ Create <link....> tags in header
      if (TopicOffset < 0 ||
          HlpTopicSeek(topic, TopicOffset) == -1)
        {
          sbprintf(&dta->sbrow, "<p>TopicOffset is beyond end of File.</p>");
          TopicOffset = 0;
          HlpTopicSeek(topic, TopicOffset);
        }
      
      /* ------------- Enumerate the Topic Contents ----------------------*/
      HlpTopicEnumerate(topic, TopicEnumerationLinkRelProcedure, (LPARAM)dta);
      dta->BrowseBackCalled = FALSE;
      
      sbprintf(&dta->sbrow, 
              "<style type=\"text/css\">\n"
              "<!--\n"
              " p { margin-bottom:0pt; margin-top:0pt; }\n"
              "//-->\n"
              "</style>\n\n");

      if (dta->keyword)
        {
          sbadd(&dta->sbrow, "<meta name=\"keywords\" content=\"");
          sbhtmlputs(&dta->sbrow, dta->keyword);
          sbadd(&dta->sbrow, "\">\n");
        }
      
      sbadd(&dta->sbrow, "</head>\n");
      sbprintf(&dta->sbrow, "<body bgcolor=\"#%6x\">\n", BgColor);
      
      //------------ Enumerate the Topic ------------------------------
      HlpTopicSeek(     topic, TopicOffset);
      sbprintf(&dta->sbrow, "<div style=\"background-color: #%6x;\"><p>", dta->BgColorNsr);
      HlpTopicEnumerate(topic, TopicEnumerationProcedure, (LPARAM)dta);

      _CloseRtfPara(dta);
      _CloseTabEmulationTable(dta);
      
      sbadd(&dta->sbrow, "</body>\n</html>\n");
      
      if (dta->FormatPrint)
        {
          sbadd(&dta->sbrow,
                 "<script language=\"JavaScript\">\n"
                 "<!--\n"
                 "  window.print();\n"
                 "-->\n"
                 "</script>");
        }

      
      sbcat(dta->sbout, &dta->sbrow);
      free(dta);
    }
} // EnumerateTopic



LPCSTR 
SimpleReadRtfCommand(LPCSTR lpszRtf, LPSTR dst, LPSTR par);


LPCSTR 
SimpleReadFontTable(LPCSTR lpszRtf, LPLOGFONT lf)
{
  int   count=0;
  char *dst;

  while (*lpszRtf && *lpszRtf!='}' && count<MAX_FONTS)
	 {
		while (*lpszRtf && *lpszRtf!='{' && *lpszRtf!='}') ++lpszRtf;
		
		if (*lpszRtf=='{')
		  {
			 while (*lpszRtf && *lpszRtf!=' ') ++lpszRtf;
			 
			 if (*lpszRtf==' ')
				{  lpszRtf++;
				for (dst=lf[count++].lfFaceName;
					  *lpszRtf && *lpszRtf!=';';
					  )
				  *dst++=*lpszRtf++;
				*dst=0;
				
				}
			 
			 if (*lpszRtf==';') lpszRtf++;
			 if (*lpszRtf=='}') lpszRtf++;
		  }
	 }
  
  return lpszRtf;
} // SimpleReadFontTable


LPCSTR 
SimpleReadColorTable(LPCSTR lpszRtf, LPRGBQUAD cl)
{
  int   count=0;
  char  szRtfCommand[100],
        szRtfPar[100];

  while (*lpszRtf && *lpszRtf!='}' && count<MAX_FONTS-1)
  {
    switch (*lpszRtf)
    {
      case ';':  { count++; ++lpszRtf; } break;
      case '\\': lpszRtf=SimpleReadRtfCommand(lpszRtf+1, szRtfCommand, szRtfPar);
                      if (!stricmp(szRtfCommand,"red"  )) cl[count].rgbRed  =(BYTE)atoi(szRtfPar);
                 else if (!stricmp(szRtfCommand,"green")) cl[count].rgbGreen=(BYTE)atoi(szRtfPar);
                 else if (!stricmp(szRtfCommand,"blue" )) cl[count].rgbBlue =(BYTE)atoi(szRtfPar);
                 break;
      default:   lpszRtf++;
    }
  }

  return lpszRtf;
} // SimpleReadColorTable


LPCSTR 
SimpleReadRtfCommand(LPCSTR lpszRtf, LPSTR szRtfCommand, LPSTR szRtfPar)
{   
  char * dst;

  for  ( dst=szRtfCommand;
			(*lpszRtf>='a' && *lpszRtf<='z')||
           (*lpszRtf>='A' && *lpszRtf<='Z')||
			  *lpszRtf=='~';
         )
	 *dst++=*lpszRtf++;
  *dst=0;
  
  for  ( dst=szRtfPar;
			(*lpszRtf>='0' && *lpszRtf<='9')||
			  *lpszRtf=='-';
         )
	 *dst++=*lpszRtf++;
  *dst=0;
  
  if (*lpszRtf && strchr("\r\n\t ", *lpszRtf))
	 ++lpszRtf;
  
  return lpszRtf;
} // SimpleReadRtfCommand


/* SimpleRTfToHtml
 *
 * This routine parses the RTF-Codes transfered by HlpEnumTopic callback
 * to find out about fonts, tables etc. The Coding only supports the
 * subset of RTF used by Help Access Library, it would fail on a
 * RTF-File created by a Word-Processing application, mainly due
 * to the fact that we haven't supported restaurtaion of font-
 * attrubutes on { } parantheses.
 */
static
void 
SimpleRtfToHtml(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszRtf)
{
  char  szRtfCommand[100];
  char  szRtfPar[100];
  
  while (*lpszRtf)
	 {
		if (*lpszRtf=='\\')
		  {
			 lpszRtf=SimpleReadRtfCommand(lpszRtf+1, szRtfCommand, szRtfPar);
		  
			 if (*lpszRtf=='}') dta->Hidden=FALSE;
			 
			 if (!stricmp(szRtfCommand,"b"))    dta->lfnow.lfWeight=szRtfPar[0]=='0'?FW_NORMAL:FW_BOLD;
			 else if (!stricmp(szRtfCommand,"v"))    dta->Hidden=szRtfPar[0]!='0';
			 else if (!stricmp(szRtfCommand,"ul"))   dta->lfnow.lfUnderline=szRtfPar[0]!='0';
			 else if (!stricmp(szRtfCommand,"i"))    dta->lfnow.lfItalic=szRtfPar[0]!='0';
			 else if (!stricmp(szRtfCommand,"fonttbl"))  lpszRtf=SimpleReadFontTable(lpszRtf, dta->lf);
			 else if (!stricmp(szRtfCommand,"colortbl")) lpszRtf=SimpleReadColorTable(lpszRtf, dta->cl);
			 else if (!stricmp(szRtfCommand,"fs"))       dta->lfnow.lfHeight=atoi(szRtfPar);
			 else if (!stricmp(szRtfCommand,"cf"))       dta->clnow=dta->cl[atoi(szRtfPar)];
			 else if (!stricmp(szRtfCommand,"f" ))       strcpy(dta->lfnow.lfFaceName, dta->lf[atoi(szRtfPar)].lfFaceName);
			 else if (!stricmp(szRtfCommand,"brdrb"))    dta->InBrdrb=TRUE;
          else if (!stricmp(szRtfCommand,"brdrt"))    dta->InBrdrt=TRUE;
			 else if (!stricmp(szRtfCommand,"brdrl"))    dta->InBrdrl=TRUE;
          else if (!stricmp(szRtfCommand,"brdrr"))    dta->InBrdrr=TRUE;
          else if (!stricmp(szRtfCommand,"box"))      dta->InBrdrl=dta->InBrdrr=dta->InBrdrt=dta->InBrdrb=TRUE;
			 else if (!stricmp(szRtfCommand,"~"))        sbadd(&dta->sbrow, "&nbsp");
			 
			 else if (!stricmp(szRtfCommand,"intbl" ))   
				{ 
				  if (!dta->HtmlTable)
					 { 
						dta->HtmlTable=TRUE;

						/* Always close font when changing table items */
                  _CloseRtfPara(dta);
                  _CloseTabEmulationTable(dta);

						sbadd(&dta->sbrow, "\n<table border=\"0\" class=\"winhelp\">\n");
					 }
				  dta->RtfTable=TRUE;
				}
			 else	if (!stricmp(szRtfCommand,"row" ))     
            { 
              /* Always close font when changing table items */
              _CloseRtfPara(dta);
              _CloseTabEmulationTable(dta);
              
              dta->RtfTable=FALSE;
              sbadd(&dta->sbrow, "</td></tr>");
              dta->HtmlRow=FALSE;
            }
          else if (!stricmp(szRtfCommand,"cell" ))    
            {
              /* Always close font when changing table items */
              _CloseRtfPara(dta);
              _CloseTabEmulationTable(dta);
              
              sbadd(&dta->sbrow, "</td><td valign=\"top\">");
            }
          else if (!stricmp(szRtfCommand,"sa"))
            {
              dta->rtf_sa = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"sl"))
            {
              dta->rtf_sl = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"sb"))
            {
              dta->rtf_sb = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"li"))
            {
              dta->rtf_li = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"ri"))
            {
              dta->rtf_ri = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"fi"))
            {
              dta->rtf_fi = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"par"))
            {
              /* End of Paragraph */
              _CloseRtfPara(dta);
            }
          else if (!stricmp(szRtfCommand, "pard"))
            {
              /* Close the currently opened paragraph */
              _CloseRtfPara(dta);

              /* Reset Paragraph settings back to the default */
              dta->InBrdrl=
              dta->InBrdrr=
              dta->InBrdrt=
                dta->InBrdrb=FALSE;

              dta->tabcurrent=
              dta->tabalign =
              dta->tabs_rtf.tabcount = 
              dta->rtf_li =
              dta->rtf_ri =
              dta->rtf_fi =
              dta->rtf_sl =
              dta->rtf_sa =
              dta->rtf_sb = 0;
              
              dta->InNsr = FALSE;
            }
          else if (!stricmp(szRtfCommand, "keepn"))
            {
              dta->InNsr = TRUE;
            }
          else if (!stricmp(szRtfCommand, "tqc"))
            {
              // Next tabstop center aligned
              dta->tabalign = 1;
            }
          else if (!stricmp(szRtfCommand, "tqr"))
            {
              // Next tabstop right aligned
              dta->tabalign = 2;
            }
          else if (!stricmp(szRtfCommand,"tx") &&
                   dta->tabs_rtf.tabcount < MAX_TABSTOPS)
            {
              // Define a tabstop
              TABSTOP *tstop = dta->tabs_rtf.tabstop + dta->tabs_rtf.tabcount++;
              tstop->tabpos = atoi(szRtfPar);
              tstop->tabalign = dta->tabalign;
              dta->tabalign = 0;
            }
			 else if (!stricmp(szRtfCommand,"tab"))
            {
              if (dta->tabcurrent <= dta->tabs_html.tabcount && dta->tabs_html.tabcount!=0)
                {
                  _OpenPara(dta);
                  _NextTab(dta);
                }
              else
                {
                  /* More tabs than we have tabstops defined for. */
                  sbadd(&dta->sbrow, "&nbsp;&nbsp;");
                }
            }
			 else if (!stricmp(szRtfCommand,"line")) 
            {
              _RtfLineCommand(dta);
            }
		  }
      else
		  {
			 if (!strchr(" \r\n\t{}",*lpszRtf) && !dta->Hidden)
				{
              _OpenPara(dta);

              szRtfCommand[0] = *lpszRtf;
				  szRtfCommand[1] = 0;
				  sbhtmlputs(&dta->sbrow, szRtfCommand);
				  lpszRtf++;
				}
			 else
				lpszRtf++;
		  }
	 }
} // SimpleRtfToHtml



/**
 * Returns TRUE if the string given in the parameter is empty.
 *
 * @param str    String to be analyzed
 *
 * @return TRUE if string contains entirely of whitespace characters
 */
static
BOOL
_IsEmptyString(const char *str)
{
  while (*str && (*str==' '||*str=='\t'||*str=='\r'||*str=='\n'))
    str++;

  return !*str;
}


/*
 * Local variables:
 *  tab-width: 3
 * End:
 */
