/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* wmf2png.c - Part of WinHelpCGI
  
   Converts a Graphic in Windows MetaFile Format into a PNG file.
  
   Interesting limitations:
   - MAPs/SHGs are not supported.


   Conversion of Windows Metafiles into PNG are very complex since MetaFiles
   are Vector Graphics and PNG files are Bitmap images.
   Most of the complex work ist done by the libwmf library.
   http://download.sourceforge.net/wvware/

   (C)1996-2004 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/

#ifdef HAVE_LIBWMF

#include <windows.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <libwmf/api.h>
#include <libwmf/gd.h>

#include "hlpacces.h"           // Help Access Library Functions
#include "shgimpor.h"           // Import of Bitmaps in Help File Format
#include "stringbuffer.h"       // STRINGBUFFER Appendable strings structures
#include "wmf2png.h"            // Our own private header file.


#ifdef _Windows
#error "don't do this..."
#endif


static int _user_write_function(void* context,char* buffer,int length);

/**
 * Create a memory block that contains the placeable metafile header
 * and the metafile data.
 */

static
BYTE *
_CombinePlacableData(HMETAFILE hMetaFile, int iImageSize, METAFILEHEADER *WMFHead)
{
  int requiredsize = iImageSize + sizeof(METAFILEHEADER);
  BYTE *block = (BYTE *) malloc(requiredsize);
  BYTE *dest = block;
  if (dest)
    {
      /* libwmf expects bytes in Intel byte order, even on non-monopolistic machines.
       * So write it bytes for byte... */
      *dest++ = (BYTE) (WMFHead->key >> 0);
      *dest++ = (BYTE) (WMFHead->key >> 8);
      *dest++ = (BYTE) (WMFHead->key >>16);
      *dest++ = (BYTE) (WMFHead->key >>24);

      *dest++ = 0; // hmf 
      *dest++ = 0;

      *dest++ = (BYTE) (WMFHead->bbox.left >> 0);
      *dest++ = (BYTE) (WMFHead->bbox.left >> 8);

      *dest++ = (BYTE) (WMFHead->bbox.top >> 0);
      *dest++ = (BYTE) (WMFHead->bbox.top >> 8);
      
      *dest++ = (BYTE) (WMFHead->bbox.right >> 0);
      *dest++ = (BYTE) (WMFHead->bbox.right >> 8);
      
      *dest++ = (BYTE) (WMFHead->bbox.bottom >> 0);
      *dest++ = (BYTE) (WMFHead->bbox.bottom >> 8);
      
      *dest++ = (BYTE) (WMFHead->inch >> 0);
      *dest++ = (BYTE) (WMFHead->inch >> 8);
      
      *dest++ = 0; // reserved
      *dest++ = 0;
      *dest++ = 0;
      *dest++ = 0;

      *dest++ = (BYTE) (WMFHead->checksum >> 0);
      *dest++ = (BYTE) (WMFHead->checksum >> 8);
      
      /* Copy the main part */
      memcpy(dest, GlobalLock((HGLOBAL) hMetaFile), iImageSize);
    }

  return block;
}




/**
 * Convert an in-memory representation of a metafile and a header-structure
 * to a PNG-File to be sent to fout.
 *
 * @param sbout               Output STRINGBUFFER
 * @param hMetaFile           In-Memory representation of the Metafile
 * @param iImageSize          Size of the memory block in hMetaFile
 * @param WMFHead             Placeable Metafile header
 */

void
Wmf2Png(STRINGBUFFER *sbout, HMETAFILE hMetaFile, int iImageSize, METAFILEHEADER *WMFHead)
{
  BYTE *placeable;

  /* First create a Version of the metafile data that contains
   * all the neccessery headers, since libwmf needs it that way.
   */
  placeable = _CombinePlacableData(hMetaFile, iImageSize, WMFHead);
  if (placeable)
    {
      wmfAPI_Options options;
      wmfAPI* API;
      
      wmf_error_t error;
      
      unsigned long flags;
      
      /* */
      
      flags = WMF_OPT_FUNCTION;
      options.function = wmf_gd_function;
      
      /* Other Options */
      
      error = wmf_api_create (&API,flags,&options);
      if (error == wmf_E_None)
        {
          error = wmf_mem_open (API,(unsigned char*) placeable, iImageSize + sizeof(METAFILEHEADER));
          if (error == wmf_E_None)
            {
              wmf_gd_t* ddata;
              wmfD_Rect bbox;

              ddata = WMF_GD_GetData (API);
              if ((ddata->flags & WMF_GD_SUPPORTS_PNG) == 0)
                {	
                  wmf_api_destroy (API);
                  fprintf (stderr,"libwmf does not support PNG conversion!\n");
                  exit(EXIT_FAILURE);
                }  

              /* */
              error = wmf_scan (API,0,&bbox);
              if (error == wmf_E_None)
                {
                  ddata->type = wmf_gd_png;
                  
                  ddata->flags |= WMF_GD_OUTPUT_MEMORY | WMF_GD_OWN_BUFFER;
                  ddata->sink.context = sbout;
                  ddata->sink.function = _user_write_function;
                  
                  ddata->bbox = bbox;

                  // Scale it to 96 DPI for typcal screen resolution
                  ddata->width  = (WMFHead->bbox.right - WMFHead->bbox.left) * 96 / WMFHead->inch;
                  ddata->height  = (WMFHead->bbox.bottom - WMFHead->bbox.top) * 96 / WMFHead->inch;
                  
                  error = wmf_play (API,0,&bbox);
                  if (error == wmf_E_None)
                    {	
                    }
                }
                
              error = wmf_mem_close (API);
            }
       }
      else
        {	
        }
      
      wmf_api_destroy (API);
      
      free(placeable);
    }
}

static int _user_write_function(void* context,char* buffer,int length)
{
  sbwrite((STRINGBUFFER *) context, buffer, length);
  return length;
}


#endif // HAVE_LIBWMF
